@@ -18,6 +18,13 @@
 #define __NR_sys_vmsplice	316
 #endif
 
+#ifndef __NR_async_exec
+# define __NR_async_exec	320
+# define __NR_async_wait	321
+# define __NR_umem_add		322
+# define __NR_async_thread	323
+#endif
+
 #define nop	__asm__ __volatile__("rep;nop": : :"memory")
 
 static inline unsigned long ffz(unsigned long bitmask)
