@@ -26,20 +26,20 @@ static void check_str_update(struct thread_data *td)
 			break;
 		case TD_RUNNING:
 			if (td_rw(td)) {
-				if (td->sequential)
-					c = 'M';
-				else
+				if (td_random(td))
 					c = 'm';
+				else
+					c = 'M';
 			} else if (td_read(td)) {
-				if (td->sequential)
-					c = 'R';
-				else
+				if (td_random(td))
 					c = 'r';
-			} else {
-				if (td->sequential)
-					c = 'W';
 				else
+					c = 'R';
+			} else {
+				if (td_random(td))
 					c = 'w';
+				else
+					c = 'W';
 			}
 			break;
 		case TD_VERIFYING:
@@ -70,7 +70,7 @@ static void check_str_update(struct thread_data *td)
 static void eta_to_str(char *str, int eta_sec)
 {
 	unsigned int d, h, m, s;
-	static int always_d, always_h;
+	int disp_hour = 0;
 
 	d = h = m = s = 0;
 
@@ -82,14 +82,13 @@ static void eta_to_str(char *str, int eta_sec)
 	eta_sec /= 24;
 	d = eta_sec;
 
-	if (d || always_d) {
-		always_d = 1;
+	if (d) {
+		disp_hour = 1;
 		str += sprintf(str, "%02ud:", d);
 	}
-	if (h || always_h) {
-		always_h = 1;
+
+	if (h || disp_hour)
 		str += sprintf(str, "%02uh:", h);
-	}
 
 	str += sprintf(str, "%02um:", m);
 	str += sprintf(str, "%02us", s);
@@ -164,6 +163,15 @@ static int thread_eta(struct thread_data *td, unsigned long elapsed)
 	return eta_sec;
 }
 
+static void calc_rate(unsigned long mtime, unsigned long long *io_bytes,
+		      unsigned long long *prev_io_bytes, unsigned int *rate)
+{
+	rate[0] = (io_bytes[0] - prev_io_bytes[0]) / mtime;
+	rate[1] = (io_bytes[1] - prev_io_bytes[1]) / mtime;
+	prev_io_bytes[0] = io_bytes[0];
+	prev_io_bytes[1] = io_bytes[1];
+}
+	
 /*
  * Print status of the jobs we know about. This includes rate estimates,
  * ETA, thread state, etc.
@@ -173,29 +181,36 @@ void print_thread_status(void)
 	unsigned long elapsed = mtime_since_genesis() / 1000;
 	int i, nr_running, nr_pending, t_rate, m_rate, *eta_secs, eta_sec;
 	struct thread_data *td;
-	char eta_str[32];
+	char eta_str[128];
 	double perc = 0.0;
-
-	static unsigned long long prev_io_bytes[2];
-	static struct timeval prev_time;
-	static unsigned int r_rate, w_rate;
 	unsigned long long io_bytes[2];
-	unsigned long mtime;
+	unsigned long rate_time, disp_time, bw_avg_time;
+	struct timeval now;
+
+	static unsigned long long rate_io_bytes[2];
+	static unsigned long long disp_io_bytes[2];
+	static struct timeval rate_prev_time, disp_prev_time;
+	static unsigned int rate[2];
 
 	if (temp_stall_ts || terse_output)
 		return;
 
-	if (!prev_io_bytes[0] && !prev_io_bytes[1])
-		fill_start_time(&prev_time);
+	if (!rate_io_bytes[0] && !rate_io_bytes[1])
+		fill_start_time(&rate_prev_time);
+	if (!disp_io_bytes[0] && !disp_io_bytes[1])
+		fill_start_time(&disp_prev_time);
 
 	eta_secs = malloc(thread_number * sizeof(int));
 	memset(eta_secs, 0, thread_number * sizeof(int));
 
 	io_bytes[0] = io_bytes[1] = 0;
 	nr_pending = nr_running = t_rate = m_rate = 0;
+	bw_avg_time = ULONG_MAX;
 	for_each_td(td, i) {
-		if (td->runstate == TD_RUNNING || td->runstate == TD_VERIFYING||
-		    td->runstate == TD_FSYNCING) {
+		if (td->bw_avg_time < bw_avg_time)
+			bw_avg_time = td->bw_avg_time;
+		if (td->runstate == TD_RUNNING || td->runstate == TD_VERIFYING
+		    || td->runstate == TD_FSYNCING) {
 			nr_running++;
 			t_rate += td->rate;
 			m_rate += td->ratemin;
@@ -234,23 +249,32 @@ void print_thread_status(void)
 		eta_to_str(eta_str, eta_sec);
 	}
 
-	mtime = mtime_since_now(&prev_time);
-	if (mtime > 1000) {
-		r_rate = (io_bytes[0] - prev_io_bytes[0]) / mtime;
-		w_rate = (io_bytes[1] - prev_io_bytes[1]) / mtime;
-		fio_gettime(&prev_time, NULL);
-		memcpy(prev_io_bytes, io_bytes, sizeof(io_bytes));
+	fio_gettime(&now, NULL);
+	rate_time = mtime_since(&rate_prev_time, &now);
+
+	if (write_bw_log && rate_time> bw_avg_time) {
+		calc_rate(rate_time, io_bytes, rate_io_bytes, rate);
+		memcpy(&rate_prev_time, &now, sizeof(now));
+		add_agg_sample(rate[DDIR_READ], DDIR_READ);
+		add_agg_sample(rate[DDIR_WRITE], DDIR_WRITE);
 	}
 
+	disp_time = mtime_since(&disp_prev_time, &now);
+	if (disp_time < 1000)
+		return;
+
+	calc_rate(disp_time, io_bytes, disp_io_bytes, rate);
+	memcpy(&disp_prev_time, &now, sizeof(now));
+
 	if (!nr_running && !nr_pending)
 		return;
 
-	printf("Threads running: %d", nr_running);
+	printf("Threads: %d", nr_running);
 	if (m_rate || t_rate)
-		printf(", commitrate %d/%dKiB/sec", t_rate, m_rate);
+		printf(", CR=%d/%d KiB/s", t_rate, m_rate);
 	if (eta_sec != INT_MAX && nr_running) {
 		perc *= 100.0;
-		printf(": [%s] [%3.2f%% done] [%6u/%6u kb/s] [eta %s]", run_str, perc, r_rate, w_rate, eta_str);
+		printf(": [%s] [%3.1f%% done] [%6u/%6u kb/s] [eta %s]", run_str, perc, rate[0], rate[1], eta_str);
 	}
 	printf("\r");
 	fflush(stdout);
