@@ -15,12 +15,18 @@ static int file_ok(struct thread_data *td, struct fio_file *f)
 {
 	struct stat st;
 
-	if (td->filetype != FIO_TYPE_FILE)
+	if (td->filetype != FIO_TYPE_FILE || (td->io_ops->flags & FIO_NULLIO))
 		return 0;
 
-	if (stat(f->file_name, &st) == -1)
+	if (lstat(f->file_name, &st) == -1)
 		return 1;
-	else if (st.st_size < (off_t) f->file_size)
+
+	/*
+	 * if it's a special file, size is always ok for now
+	 */
+	if (S_ISCHR(st.st_mode) || S_ISBLK(st.st_mode))
+		return 0;
+	if (st.st_size < (off_t) f->file_size)
 		return 1;
 
 	return 0;
@@ -35,17 +41,17 @@ static int create_file(struct thread_data *td, struct fio_file *f)
 
 	f->fd = open(f->file_name, O_WRONLY | O_CREAT | O_TRUNC, 0644);
 	if (f->fd < 0) {
-		td_verror(td, errno);
+		td_verror(td, errno, "open");
 		return 1;
 	}
 
 	if (ftruncate(f->fd, f->file_size) == -1) {
-		td_verror(td, errno);
+		td_verror(td, errno, "ftruncate");
 		goto err;
 	}
 
 	if (posix_fallocate(f->fd, 0, f->file_size) < 0) {
-		td_verror(td, errno);
+		td_verror(td, errno, "posix_fallocate");
 		goto err;
 	}
 
@@ -65,9 +71,9 @@ static int create_file(struct thread_data *td, struct fio_file *f)
 			continue;
 		} else {
 			if (r < 0)
-				td_verror(td, errno);
+				td_verror(td, errno, "write");
 			else
-				td_verror(td, EIO);
+				td_verror(td, EIO, "write");
 
 			break;
 		}
@@ -103,16 +109,25 @@ static int create_files(struct thread_data *td)
 		return 0;
 
 	need_create = 0;
-	if (td->filetype == FIO_TYPE_FILE)
-		for_each_file(td, f, i)
-			need_create += file_ok(td, f);
+	if (td->filetype == FIO_TYPE_FILE) {
+		for_each_file(td, f, i) {
+			int file_there = !file_ok(td, f);
+
+			if (file_there && td_write(td) && !td->overwrite) {
+				unlink(f->file_name);
+				file_there = 0;
+			}
+
+			need_create += !file_there;
+		}
+	}
 
 	if (!need_create)
 		return 0;
 
 	if (!td->total_file_size) {
 		log_err("Need size for create\n");
-		td_verror(td, EINVAL);
+		td_verror(td, EINVAL, "file_size");
 		return 1;
 	}
 
@@ -144,9 +159,19 @@ static int file_size(struct thread_data *td, struct fio_file *f)
 {
 	struct stat st;
 
+	/*
+	 * if we are not doing real io, just pretend the file is as large
+	 * as the size= given. this works fine with nrfiles > 1 as well,
+	 * we only really care about it being at least as big as size=
+	 */
+	if (td->io_ops->flags & FIO_NULLIO) {
+		f->real_file_size = f->file_size = td->total_file_size;
+		return 0;
+	}
+
 	if (td->overwrite) {
 		if (fstat(f->fd, &st) == -1) {
-			td_verror(td, errno);
+			td_verror(td, errno, "fstat");
 			return 1;
 		}
 
@@ -167,7 +192,7 @@ static int bdev_size(struct thread_data *td, struct fio_file *f)
 
 	r = blockdev_size(f->fd, &bytes);
 	if (r) {
-		td_verror(td, r);
+		td_verror(td, r, "blockdev_size");
 		return 1;
 	}
 
@@ -214,19 +239,21 @@ int file_invalidate_cache(struct thread_data *td, struct fio_file *f)
 	 */
 	if (td->io_ops->flags & FIO_MMAPIO)
 		ret = madvise(f->mmap, f->file_size, MADV_DONTNEED);
-	else if (td->filetype == FIO_TYPE_FILE)
-		ret = fadvise(f->fd, f->file_offset, f->file_size, POSIX_FADV_DONTNEED);
-	else if (td->filetype == FIO_TYPE_BD)
-		ret = blockdev_invalidate_cache(f->fd);
-	else if (td->filetype == FIO_TYPE_CHAR)
+	else if (td->filetype == FIO_TYPE_FILE) {
+		if (!td->odirect)
+			ret = fadvise(f->fd, f->file_offset, f->file_size, POSIX_FADV_DONTNEED);
+	} else if (td->filetype == FIO_TYPE_BD) {
+		if (!td->odirect)
+			ret = blockdev_invalidate_cache(f->fd);
+	} else if (td->filetype == FIO_TYPE_CHAR)
 		ret = 0;
 
 	if (ret < 0) {
-		td_verror(td, errno);
+		td_verror(td, errno, "invalidate_cache");
 		return 1;
 	}
 
-	return 0;
+	return ret;
 }
 
 static int __setup_file_mmap(struct thread_data *td, struct fio_file *f)
@@ -246,21 +273,21 @@ static int __setup_file_mmap(struct thread_data *td, struct fio_file *f)
 	f->mmap = mmap(NULL, f->file_size, flags, MAP_SHARED, f->fd, f->file_offset);
 	if (f->mmap == MAP_FAILED) {
 		f->mmap = NULL;
-		td_verror(td, errno);
+		td_verror(td, errno, "mmap");
 		return 1;
 	}
 
 	if (td->invalidate_cache && file_invalidate_cache(td, f))
 		return 1;
 
-	if (td->sequential) {
+	if (!td_random(td)) {
 		if (madvise(f->mmap, f->file_size, MADV_SEQUENTIAL) < 0) {
-			td_verror(td, errno);
+			td_verror(td, errno, "madvise");
 			return 1;
 		}
 	} else {
 		if (madvise(f->mmap, f->file_size, MADV_RANDOM) < 0) {
-			td_verror(td, errno);
+			td_verror(td, errno, "madvise");
 			return 1;
 		}
 	}
@@ -287,14 +314,14 @@ static int __setup_file_plain(struct thread_data *td, struct fio_file *f)
 	if (td->invalidate_cache && file_invalidate_cache(td, f))
 		return 1;
 
-	if (td->sequential) {
+	if (!td_random(td)) {
 		if (fadvise(f->fd, f->file_offset, f->file_size, POSIX_FADV_SEQUENTIAL) < 0) {
-			td_verror(td, errno);
+			td_verror(td, errno, "fadvise");
 			return 1;
 		}
 	} else {
 		if (fadvise(f->fd, f->file_offset, f->file_size, POSIX_FADV_RANDOM) < 0) {
-			td_verror(td, errno);
+			td_verror(td, errno, "fadvise");
 			return 1;
 		}
 	}
@@ -320,33 +347,48 @@ static int setup_file(struct thread_data *td, struct fio_file *f)
 {
 	int flags = 0;
 
-	if (td->odirect)
-		flags |= OS_O_DIRECT;
-	if (td->sync_io)
-		flags |= O_SYNC;
+	if (td->io_ops->flags & FIO_NETIO)
+		return 0;
 
-	if (td_write(td) || td_rw(td)) {
-		flags |= O_RDWR;
+	/*
+	 * we need a valid file descriptor, but don't create a real file.
+	 * lets just dup stdout, seems like a sensible approach.
+	 */
+	if (td->io_ops->flags & FIO_NULLIO)
+		f->fd = dup(STDOUT_FILENO);
+	else {
+		if (td->odirect)
+			flags |= OS_O_DIRECT;
+		if (td->sync_io)
+			flags |= O_SYNC;
 
-		if (td->filetype == FIO_TYPE_FILE) {
-			if (!td->overwrite)
-				flags |= O_TRUNC;
-
-			flags |= O_CREAT;
-		}
-
-		f->fd = open(f->file_name, flags, 0600);
-	} else {
-		if (td->filetype == FIO_TYPE_CHAR)
+		if (td_write(td) || td_rw(td)) {
 			flags |= O_RDWR;
-		else
-			flags |= O_RDONLY;
 
-		f->fd = open(f->file_name, flags);
+			if (td->filetype == FIO_TYPE_FILE) {
+				if (!td->overwrite)
+					flags |= O_TRUNC;
+
+				flags |= O_CREAT;
+			}
+
+			f->fd = open(f->file_name, flags, 0600);
+		} else {
+			if (td->filetype == FIO_TYPE_CHAR)
+				flags |= O_RDWR;
+			else
+				flags |= O_RDONLY;
+
+			f->fd = open(f->file_name, flags);
+		}
 	}
 
 	if (f->fd == -1) {
-		td_verror(td, errno);
+		int __e = errno;
+
+		td_verror(td, __e, "open");
+		if (__e == EINVAL && td->odirect)
+			log_err("fio: destination does not support O_DIRECT\n");
 		return 1;
 	}
 
@@ -409,7 +451,7 @@ int setup_files(struct thread_data *td)
 	td->io_size = td->total_file_size;
 	if (td->io_size == 0) {
 		log_err("%s: no io blocks\n", td->name);
-		td_verror(td, EINVAL);
+		td_verror(td, EINVAL, "total_file_size");
 		return 1;
 	}
 
