@@ -18,40 +18,80 @@
 #include "arch.h"
 #include "os.h"
 
+#ifdef FIO_HAVE_SYSLET
+#include "syslet.h"
+#endif
+
 enum fio_ddir {
 	DDIR_READ = 0,
 	DDIR_WRITE,
 	DDIR_SYNC,
 };
 
+enum td_ddir {
+	TD_DDIR_READ		= 1 << 0,
+	TD_DDIR_WRITE		= 1 << 1,
+	TD_DDIR_RAND		= 1 << 2,
+	TD_DDIR_RW		= TD_DDIR_READ | TD_DDIR_WRITE,
+	TD_DDIR_RANDREAD	= TD_DDIR_READ | TD_DDIR_RAND,
+	TD_DDIR_RANDWRITE	= TD_DDIR_WRITE | TD_DDIR_READ,
+	TD_DDIR_RANDRW		= TD_DDIR_RW | TD_DDIR_RAND,
+};
+
+/*
+ * Use for maintaining statistics
+ */
 struct io_stat {
-	unsigned long long val;
-	unsigned long long val_sq;
 	unsigned long max_val;
 	unsigned long min_val;
 	unsigned long samples;
+
+	double mean;
+	double S;
 };
 
+/*
+ * A single data sample
+ */
 struct io_sample {
 	unsigned long time;
 	unsigned long val;
 	enum fio_ddir ddir;
 };
 
+/*
+ * Dynamically growing data sample log
+ */
 struct io_log {
 	unsigned long nr_samples;
 	unsigned long max_samples;
 	struct io_sample *log;
 };
 
+/*
+ * When logging io actions, this matches a single sent io_u
+ */
 struct io_piece {
 	struct list_head list;
 	struct fio_file *file;
 	unsigned long long offset;
-	unsigned int len;
+	unsigned long len;
 	enum fio_ddir ddir;
 };
 
+#ifdef FIO_HAVE_SYSLET
+struct syslet_req {
+	struct syslet_uatom atom;	/* the atom to submit */
+	struct syslet_uatom *head;	/* head of the sequence */
+	long ret;			/* syscall return value */
+};
+#endif
+
+enum {
+	IO_U_F_FREE	= 1 << 0,
+	IO_U_F_FLIGHT	= 1 << 1,
+};
+
 /*
  * The io unit
  */
@@ -65,15 +105,28 @@ struct io_u {
 #endif
 #ifdef FIO_HAVE_SGIO
 		struct sg_io_hdr hdr;
+#endif
+#ifdef FIO_HAVE_SYSLET
+		struct syslet_req req;
 #endif
 	};
 	struct timeval start_time;
 	struct timeval issue_time;
 
+	/*
+	 * Allocated/set buffer and length
+	 */
 	void *buf;
-	unsigned int buflen;
+	unsigned long buflen;
 	unsigned long long offset;
 
+	/*
+	 * IO engine state, may be different from above when we get
+	 * partial transfers / residual data counts
+	 */
+	void *xfer_buf;
+	unsigned long xfer_buflen;
+
 	unsigned int resid;
 	unsigned int error;
 
@@ -87,19 +140,38 @@ struct io_u {
 		unsigned int seen;
 	};
 
+	unsigned int flags;
+
 	struct fio_file *file;
 
 	struct list_head list;
+
+	/*
+	 * Callback for io completion
+	 */
+	int (*end_io)(struct io_u *);
+};
+
+/*
+ * io_ops->queue() return values
+ */
+enum {
+	FIO_Q_COMPLETED	= 0,		/* completed sync */
+	FIO_Q_QUEUED	= 1,		/* queued, will complete async */
+	FIO_Q_BUSY	= 2,		/* no more room, call ->commit() */
 };
 
 #define FIO_HDR_MAGIC	0xf00baaef
 
 enum {
-	VERIFY_NONE = 0,
-	VERIFY_MD5,
-	VERIFY_CRC32,
+	VERIFY_NONE = 0,		/* no verification */
+	VERIFY_MD5,			/* md5 sum data blocks */
+	VERIFY_CRC32,			/* crc32 sum data blocks */
 };
 
+/*
+ * A header structure associated with each checksummed data block
+ */
 struct verify_header {
 	unsigned int fio_magic;
 	unsigned int len;
@@ -132,18 +204,24 @@ enum fio_memtype {
  * The type of object we are working on
  */
 enum fio_filetype {
-	FIO_TYPE_FILE = 1,
-	FIO_TYPE_BD,
-	FIO_TYPE_CHAR,
+	FIO_TYPE_FILE = 1,		/* plain file */
+	FIO_TYPE_BD,			/* block device */
+	FIO_TYPE_CHAR,			/* character device */
 };
 
 enum fio_ioengine_flags {
-	FIO_SYNCIO	= 1 << 0,
-	FIO_CPUIO	= 1 << 1,
-	FIO_MMAPIO	= 1 << 2,
-	FIO_RAWIO	= 1 << 3,
+	FIO_SYNCIO	= 1 << 0,	/* io engine has synchronous ->queue */
+	FIO_CPUIO	= 1 << 1,	/* cpu burner, doesn't do real io */
+	FIO_MMAPIO	= 1 << 2,	/* uses memory mapped io */
+	FIO_RAWIO	= 1 << 3,	/* some sort of direct/raw io */
+	FIO_NETIO	= 1 << 4,	/* networked io */
+	FIO_NULLIO	= 1 << 5,	/* no real data transfer (cpu/null) */
 };
 
+/*
+ * Each thread_data structure has a number of files associated with it,
+ * this structure holds state information for a single file.
+ */
 struct fio_file {
 	/*
 	 * A file may not be a file descriptor, let the io engine decide
@@ -160,16 +238,53 @@ struct fio_file {
 	unsigned long long last_pos;
 	unsigned long long last_completed_pos;
 
+	/*
+	 * block map for random io
+	 */
 	unsigned long *file_map;
 	unsigned int num_maps;
+	unsigned int last_free_lookup;
 
 	unsigned int unlink;
 };
 
+struct thread_stat {
+	struct io_log *slat_log;
+	struct io_log *clat_log;
+	struct io_log *bw_log;
+
+	/*
+	 * bandwidth and latency stats
+	 */
+	struct io_stat clat_stat[2];		/* completion latency */
+	struct io_stat slat_stat[2];		/* submission latency */
+	struct io_stat bw_stat[2];		/* bandwidth stats */
+
+	unsigned long long stat_io_bytes[2];
+	struct timeval stat_sample_time[2];
+
+	/*
+	 * fio system usage accounting
+	 */
+	struct rusage ru_start;
+	struct rusage ru_end;
+	unsigned long usr_time;
+	unsigned long sys_time;
+	unsigned long ctx;
+};
+
+/*
+ * How many depth levels to log
+ */
+#define FIO_IO_U_MAP_NR	8
+#define FIO_IO_U_LAT_NR 12
+
 /*
  * This describes a single thread/process executing a fio job.
  */
 struct thread_data {
+	int pad;
+	char *description;
 	char *name;
 	char *directory;
 	char *filename;
@@ -177,23 +292,25 @@ struct thread_data {
 	pthread_t thread;
 	int thread_number;
 	int groupid;
+	struct thread_stat ts;
 	enum fio_filetype filetype;
 	struct fio_file *files;
 	unsigned int nr_files;
 	unsigned int nr_uniq_files;
-	unsigned int next_file;
+	union {
+		unsigned int next_file;
+		os_random_state_t next_file_state;
+	};
 	int error;
 	pid_t pid;
 	char *orig_buffer;
 	size_t orig_buffer_size;
 	volatile int terminate;
 	volatile int runstate;
-	enum fio_ddir ddir;
-	unsigned int iomix;
+	enum td_ddir td_ddir;
 	unsigned int ioprio;
 	unsigned int last_was_sync;
 
-	unsigned int sequential;
 	unsigned int odirect;
 	unsigned int invalidate_cache;
 	unsigned int create_serialize;
@@ -217,6 +334,7 @@ struct thread_data {
 	unsigned int hugepage_size;
 	unsigned int rw_min_bs;
 	unsigned int thinktime;
+	unsigned int thinktime_spin;
 	unsigned int thinktime_blocks;
 	unsigned int fsync_blocks;
 	unsigned int start_delay;
@@ -232,6 +350,8 @@ struct thread_data {
 	unsigned int stonewall;
 	unsigned int numjobs;
 	unsigned int iodepth;
+	unsigned int iodepth_low;
+	unsigned int iodepth_batch;
 	os_cpu_mask_t cpumask;
 	unsigned int iolog;
 	unsigned int read_iolog;
@@ -239,6 +359,7 @@ struct thread_data {
 	unsigned int rwmixread;
 	unsigned int rwmixwrite;
 	unsigned int nice;
+	unsigned int file_service_type;
 
 	char *read_iolog_file;
 	char *write_iolog_file;
@@ -263,8 +384,13 @@ struct thread_data {
 	 * Current IO depth and list of free and busy io_u's.
 	 */
 	unsigned int cur_depth;
+	unsigned int io_u_map[FIO_IO_U_MAP_NR];
+	unsigned int io_u_lat[FIO_IO_U_LAT_NR];
+	unsigned long total_io_u;
 	struct list_head io_u_freelist;
 	struct list_head io_u_busylist;
+	struct list_head io_u_requeues;
+	unsigned int io_u_queued;
 
 	/*
 	 * Rate state
@@ -283,10 +409,11 @@ struct thread_data {
 	unsigned long long start_offset;
 	unsigned long long total_io_size;
 
+	unsigned long io_issues[2];
 	unsigned long long io_blocks[2];
 	unsigned long long io_bytes[2];
-	unsigned long long zone_bytes;
 	unsigned long long this_io_bytes[2];
+	unsigned long long zone_bytes;
 	volatile int mutex;
 
 	/*
@@ -300,33 +427,10 @@ struct thread_data {
 	unsigned int cpuload;
 	unsigned int cpucycle;
 
-	/*
-	 * bandwidth and latency stats
-	 */
-	struct io_stat clat_stat[2];		/* completion latency */
-	struct io_stat slat_stat[2];		/* submission latency */
-	struct io_stat bw_stat[2];		/* bandwidth stats */
-
-	unsigned long long stat_io_bytes[2];
-	struct timeval stat_sample_time[2];
-
-	struct io_log *slat_log;
-	struct io_log *clat_log;
-	struct io_log *bw_log;
-
 	struct timeval start;	/* start of this loop */
 	struct timeval epoch;	/* time job was started */
 	struct timeval end_time;/* time job ended */
 
-	/*
-	 * fio system usage accounting
-	 */
-	struct rusage ru_start;
-	struct rusage ru_end;
-	unsigned long usr_time;
-	unsigned long sys_time;
-	unsigned long ctx;
-
 	/*
 	 * read/write mixed workload state
 	 */
@@ -345,18 +449,43 @@ struct thread_data {
 	 */
 	struct list_head io_hist_list;
 	struct list_head io_log_list;
+
+	/*
+	 * timeout handling
+	 */
+	struct timeval timeout_end;
+	struct itimerval timer;
 };
 
-#define __td_verror(td, err, msg)					\
+/*
+ * roundrobin available files, or choose one at random.
+ */
+enum {
+	FIO_FSERVICE_RANDOM	= 1,
+	FIO_FSERVICE_RR		= 2,
+};
+
+/*
+ * 30 second per-io_u timeout, with 5 second intervals to avoid resetting
+ * the timer on each queue operation.
+ */
+#define IO_U_TIMEOUT_INC	5
+#define IO_U_TIMEOUT		30
+
+#define __td_verror(td, err, msg, func)					\
 	do {								\
+		if ((td)->error)					\
+			break;						\
 		int e = (err);						\
 		(td)->error = e;					\
-		snprintf(td->verror, sizeof(td->verror) - 1, "file:%s:%d, error=%s", __FILE__, __LINE__, (msg));	\
+		snprintf(td->verror, sizeof(td->verror) - 1, "file:%s:%d, func=%s, error=%s", __FILE__, __LINE__, (func), (msg));	\
 	} while (0)
 
 
-#define td_verror(td, err)	__td_verror((td), (err), strerror((err)))
-#define td_vmsg(td, err, msg)	__td_verror((td), (err), (msg))
+#define td_verror(td, err, func)	\
+	__td_verror((td), (err), strerror((err)), (func))
+#define td_vmsg(td, err, msg, func)	\
+	__td_verror((td), (err), (msg), (func))
 
 extern int exitall_on_terminate;
 extern int thread_number;
@@ -370,9 +499,10 @@ extern unsigned long long mlock_size;
 
 extern struct thread_data *threads;
 
-#define td_read(td)		((td)->ddir == DDIR_READ)
-#define td_write(td)		((td)->ddir == DDIR_WRITE)
-#define td_rw(td)		((td)->iomix != 0)
+#define td_read(td)		((td)->td_ddir & TD_DDIR_READ)
+#define td_write(td)		((td)->td_ddir & TD_DDIR_WRITE)
+#define td_rw(td)		(((td)->td_ddir & TD_DDIR_RW) == TD_DDIR_RW)
+#define td_random(td)		((td)->td_ddir & TD_DDIR_RAND)
 
 #define BLOCKS_PER_MAP		(8 * sizeof(long))
 #define TO_MAP_BLOCK(td, f, b)	((b) - ((f)->file_offset / (td)->rw_min_bs))
@@ -393,6 +523,9 @@ static inline int should_fsync(struct thread_data *td)
 	return 0;
 }
 
+/*
+ * Disk utils as read in /sys/block/<dev>/stat
+ */
 struct disk_util_stat {
 	unsigned ios[2];
 	unsigned merges[2];
@@ -402,6 +535,9 @@ struct disk_util_stat {
 	unsigned time_in_queue;
 };
 
+/*
+ * Per-device disk util management
+ */
 struct disk_util {
 	struct list_head list;
 
@@ -416,14 +552,6 @@ struct disk_util {
 	struct timeval time;
 };
 
-struct io_completion_data {
-	int nr;				/* input */
-
-	int error;			/* output */
-	unsigned long bytes_done[2];	/* output */
-	struct timeval time;		/* output */
-};
-
 #define DISK_UTIL_MSEC	(250)
 
 #ifndef min
@@ -436,9 +564,9 @@ struct io_completion_data {
 /*
  * Log exports
  */
-extern int read_iolog_get(struct thread_data *, struct io_u *);
+extern int __must_check read_iolog_get(struct thread_data *, struct io_u *);
 extern void write_iolog_put(struct thread_data *, struct io_u *);
-extern int init_iolog(struct thread_data *td);
+extern int __must_check init_iolog(struct thread_data *td);
 extern void log_io_piece(struct thread_data *, struct io_u *);
 extern void prune_io_piece_log(struct thread_data *);
 extern void write_iolog_close(struct thread_data *);
@@ -456,7 +584,11 @@ extern void update_io_ticks(void);
 extern void disk_util_timer_arm(void);
 extern void setup_log(struct io_log **);
 extern void finish_log(struct thread_data *, struct io_log *, const char *);
+extern void __finish_log(struct io_log *, const char *);
 extern int setup_rate(struct thread_data *);
+extern struct io_log *agg_io_log[2];
+extern int write_bw_log;
+extern void add_agg_sample(unsigned long, enum fio_ddir);
 
 /*
  * Time functions
@@ -469,23 +601,24 @@ extern unsigned long time_since_now(struct timeval *);
 extern unsigned long mtime_since_genesis(void);
 extern void __usec_sleep(unsigned int);
 extern void usec_sleep(struct thread_data *, unsigned long);
-extern void rate_throttle(struct thread_data *, unsigned long, unsigned int, int);
+extern void rate_throttle(struct thread_data *, unsigned long, unsigned int);
 extern void fill_start_time(struct timeval *);
 extern void fio_gettime(struct timeval *, void *);
+extern void set_genesis_time(void);
 
 /*
  * Init functions
  */
-extern int parse_options(int, char **);
-extern int init_random_state(struct thread_data *);
+extern int __must_check parse_options(int, char **);
+extern int __must_check init_random_state(struct thread_data *);
 
 /*
  * File setup/shutdown
  */
 extern void close_files(struct thread_data *);
-extern int setup_files(struct thread_data *);
-extern int open_files(struct thread_data *);
-extern int file_invalidate_cache(struct thread_data *, struct fio_file *);
+extern int __must_check setup_files(struct thread_data *);
+extern int __must_check open_files(struct thread_data *);
+extern int __must_check file_invalidate_cache(struct thread_data *, struct fio_file *);
 
 /*
  * ETA/status stuff
@@ -514,15 +647,15 @@ enum {
  * Verify helpers
  */
 extern void populate_verify_io_u(struct thread_data *, struct io_u *);
-extern int get_next_verify(struct thread_data *td, struct io_u *);
-extern int do_io_u_verify(struct thread_data *, struct io_u **);
+extern int __must_check get_next_verify(struct thread_data *td, struct io_u *);
+extern int __must_check verify_io_u(struct io_u *);
 
 /*
  * Memory helpers
  */
-extern int fio_pin_memory(void);
+extern int __must_check fio_pin_memory(void);
 extern void fio_unpin_memory(void);
-extern int allocate_io_mem(struct thread_data *);
+extern int __must_check allocate_io_mem(struct thread_data *);
 extern void free_io_mem(struct thread_data *);
 
 /*
@@ -530,19 +663,24 @@ extern void free_io_mem(struct thread_data *);
  */
 #define queue_full(td)	list_empty(&(td)->io_u_freelist)
 extern struct io_u *__get_io_u(struct thread_data *);
-extern struct io_u *get_io_u(struct thread_data *, struct fio_file *);
+extern struct io_u *get_io_u(struct thread_data *);
 extern void put_io_u(struct thread_data *, struct io_u *);
-extern void ios_completed(struct thread_data *, struct io_completion_data *);
-extern void io_completed(struct thread_data *, struct io_u *, struct io_completion_data *);
+extern void requeue_io_u(struct thread_data *, struct io_u **);
+extern long __must_check io_u_sync_complete(struct thread_data *, struct io_u *);
+extern long __must_check io_u_queued_complete(struct thread_data *, int);
+extern void io_u_queued(struct thread_data *, struct io_u *);
+extern void io_u_init_timeout(void);
+extern void io_u_set_timeout(struct thread_data *);
 
 /*
  * io engine entry points
  */
-extern int td_io_init(struct thread_data *);
-extern int td_io_prep(struct thread_data *, struct io_u *);
-extern int td_io_queue(struct thread_data *, struct io_u *);
-extern int td_io_sync(struct thread_data *, struct fio_file *);
-extern int td_io_getevents(struct thread_data *, int, int, struct timespec *);
+extern int __must_check td_io_init(struct thread_data *);
+extern int __must_check td_io_prep(struct thread_data *, struct io_u *);
+extern int __must_check td_io_queue(struct thread_data *, struct io_u *);
+extern int __must_check td_io_sync(struct thread_data *, struct fio_file *);
+extern int __must_check td_io_getevents(struct thread_data *, int, int, struct timespec *);
+extern int __must_check td_io_commit(struct thread_data *);
 
 /*
  * This is a pretty crappy semaphore implementation, but with the use that fio
@@ -586,18 +724,20 @@ struct ioengine_ops {
 	int (*init)(struct thread_data *);
 	int (*prep)(struct thread_data *, struct io_u *);
 	int (*queue)(struct thread_data *, struct io_u *);
+	int (*commit)(struct thread_data *);
 	int (*getevents)(struct thread_data *, int, int, struct timespec *);
 	struct io_u *(*event)(struct thread_data *, int);
 	int (*cancel)(struct thread_data *, struct io_u *);
 	void (*cleanup)(struct thread_data *);
 	void *data;
 	void *dlhandle;
+	unsigned long priv;
 };
 
-#define FIO_IOOPS_VERSION	3
+#define FIO_IOOPS_VERSION	5
 
 extern struct ioengine_ops *load_ioengine(struct thread_data *, const char *);
-extern int register_ioengine(struct ioengine_ops *);
+extern void register_ioengine(struct ioengine_ops *);
 extern void unregister_ioengine(struct ioengine_ops *);
 extern void close_ioengine(struct thread_data *);
 
@@ -613,4 +753,14 @@ extern void close_ioengine(struct thread_data *);
 #define for_each_file(td, f, i)	\
 	for ((i) = 0, (f) = &(td)->files[0]; (i) < (int) (td)->nr_files; (i)++, (f)++)
 
+#define fio_assert(td, cond)	do {	\
+	if (!(cond)) {			\
+		int *__foo = NULL;	\
+		fprintf(stderr, "file:%s:%d, assert %s failed\n", __FILE__, __LINE__, #cond);	\
+		(td)->runstate = TD_EXITED;	\
+		(td)->error = EFAULT;		\
+		*__foo = 0;			\
+	}	\
+} while (0)
+
 #endif
