@@ -18,49 +18,12 @@
 #include "fio.h"
 #include "parse.h"
 
-/*
- * The default options
- */
-#define DEF_BS			(4096)
-#define DEF_TIMEOUT		(0)
-#define DEF_RATE_CYCLE		(1000)
-#define DEF_ODIRECT		(1)
-#define DEF_IO_ENGINE		(FIO_SYNCIO)
-#define DEF_IO_ENGINE_NAME	"sync"
-#define DEF_SEQUENTIAL		(1)
-#define DEF_RAND_REPEAT		(1)
-#define DEF_OVERWRITE		(0)
-#define DEF_INVALIDATE		(1)
-#define DEF_SYNCIO		(0)
-#define DEF_RANDSEED		(0xb1899bedUL)
-#define DEF_BWAVGTIME		(500)
-#define DEF_CREATE_SER		(1)
-#define DEF_CREATE_FSYNC	(1)
-#define DEF_LOOPS		(1)
-#define DEF_VERIFY		(0)
-#define DEF_STONEWALL		(0)
-#define DEF_NUMJOBS		(1)
-#define DEF_USE_THREAD		(0)
-#define DEF_FILE_SIZE		(1024 * 1024 * 1024UL)
-#define DEF_ZONE_SIZE		(0)
-#define DEF_ZONE_SKIP		(0)
-#define DEF_RWMIX_CYCLE		(500)
-#define DEF_RWMIX_READ		(50)
-#define DEF_NICE		(0)
-#define DEF_NR_FILES		(1)
-#define DEF_UNLINK		(1)
-#define DEF_WRITE_BW_LOG	(0)
-#define DEF_WRITE_LAT_LOG	(0)
-#define DEF_NO_RAND_MAP		(0)
-#define DEF_HUGEPAGE_SIZE	FIO_HUGE_PAGE
-#define DEF_THINKTIME_BLOCKS	(1)
+#define FIO_RANDSEED		(0xb1899bedUL)
 
 #define td_var_offset(var)	((size_t) &((struct thread_data *)0)->var)
 
-static int str_rw_cb(void *, const char *);
 static int str_ioengine_cb(void *, const char *);
 static int str_mem_cb(void *, const char *);
-static int str_verify_cb(void *, const char *);
 static int str_lockmem_cb(void *, unsigned long *);
 #ifdef FIO_HAVE_IOPRIO
 static int str_prio_cb(void *, unsigned int *);
@@ -69,102 +32,111 @@ static int str_prioclass_cb(void *, unsigned int *);
 static int str_exitall_cb(void);
 static int str_cpumask_cb(void *, unsigned int *);
 
+#define __stringify_1(x)	#x
+#define __stringify(x)		__stringify_1(x)
+
 /*
  * Map of job/command line options
  */
 static struct fio_option options[] = {
+	{
+		.name	= "description",
+		.type	= FIO_OPT_STR_STORE,
+		.off1	= td_var_offset(description),
+		.help	= "Text job description",
+	},
 	{
 		.name	= "name",
 		.type	= FIO_OPT_STR_STORE,
 		.off1	= td_var_offset(name),
+		.help	= "Name of this job",
 	},
 	{
 		.name	= "directory",
 		.type	= FIO_OPT_STR_STORE,
 		.off1	= td_var_offset(directory),
+		.help	= "Directory to store files in",
 	},
 	{
 		.name	= "filename",
 		.type	= FIO_OPT_STR_STORE,
 		.off1	= td_var_offset(filename),
+		.help	= "Force the use of a specific file",
 	},
 	{
 		.name	= "rw",
 		.type	= FIO_OPT_STR,
-		.cb	= str_rw_cb,
+		.off1	= td_var_offset(td_ddir),
+		.help	= "IO direction",
+		.def	= "read",
+		.posval = {
+			  { .ival = "read", .oval = TD_DDIR_READ },
+			  { .ival = "write", .oval = TD_DDIR_WRITE },
+			  { .ival = "randread", .oval = TD_DDIR_RANDREAD },
+			  { .ival = "randwrite", .oval = TD_DDIR_RANDWRITE },
+			  { .ival = "rw", .oval = TD_DDIR_RW },
+			  { .ival = "randrw", .oval = TD_DDIR_RANDRW },
+			  },
 	},
 	{
 		.name	= "ioengine",
 		.type	= FIO_OPT_STR,
 		.cb	= str_ioengine_cb,
-	},
-	{
-		.name	= "mem",
-		.type	= FIO_OPT_STR,
-		.cb	= str_mem_cb,
-	},
-	{
-		.name	= "verify",
-		.type	= FIO_OPT_STR,
-		.cb	= str_verify_cb,
-	},
-	{
-		.name	= "write_iolog",
-		.type	= FIO_OPT_STR_STORE,
-		.off1	= td_var_offset(write_iolog_file),
-	},
-	{
-		.name	= "read_iolog",
-		.type	= FIO_OPT_STR_STORE,
-		.off1	= td_var_offset(read_iolog_file),
-	},
-	{
-		.name	= "exec_prerun",
-		.type	= FIO_OPT_STR_STORE,
-		.off1	= td_var_offset(exec_prerun),
-	},
-	{
-		.name	= "exec_postrun",
-		.type	= FIO_OPT_STR_STORE,
-		.off1	= td_var_offset(exec_postrun),
-	},
-#ifdef FIO_HAVE_IOSCHED_SWITCH
-	{
-		.name	= "ioscheduler",
-		.type	= FIO_OPT_STR_STORE,
-		.off1	= td_var_offset(ioscheduler),
-	},
+		.help	= "IO engine to use",
+		.def	= "sync",
+		.posval	= {
+			  { .ival = "sync", },
+#ifdef FIO_HAVE_LIBAIO
+			  { .ival = "libaio", },
 #endif
+#ifdef FIO_HAVE_POSIXAIO
+			  { .ival = "posixaio", },
+#endif
+			  { .ival = "mmap", },
+#ifdef FIO_HAVE_SPLICE
+			  { .ival = "splice", },
+#endif
+#ifdef FIO_HAVE_SGIO
+			  { .ival = "sg", },
+#endif
+			  { .ival = "null", }, { .ival = "net", },
+#ifdef FIO_HAVE_SYSLET
+			  { .ival = "syslet-rw", },
+#endif
+			  },
+	},
+	{
+		.name	= "iodepth",
+		.type	= FIO_OPT_INT,
+		.off1	= td_var_offset(iodepth),
+		.help	= "Amount of IO buffers to keep in flight",
+		.def	= "1",
+	},
+	{
+		.name	= "iodepth_batch",
+		.type	= FIO_OPT_INT,
+		.off1	= td_var_offset(iodepth_batch),
+		.help	= "Number of IO to submit in one go",
+	},
+	{
+		.name	= "iodepth_low",
+		.type	= FIO_OPT_INT,
+		.off1	= td_var_offset(iodepth_low),
+		.help	= "Low water mark for queuing depth",
+	},
 	{
 		.name	= "size",
 		.type	= FIO_OPT_STR_VAL,
 		.off1	= td_var_offset(total_file_size),
+		.help	= "Size of device or file",
 	},
 	{
 		.name	= "bs",
 		.type	= FIO_OPT_STR_VAL_INT,
 		.off1	= td_var_offset(bs[DDIR_READ]),
 		.off2	= td_var_offset(bs[DDIR_WRITE]),
-	},
-	{
-		.name	= "offset",
-		.type	= FIO_OPT_STR_VAL,
-		.off1	= td_var_offset(start_offset),
-	},
-	{
-		.name	= "zonesize",
-		.type	= FIO_OPT_STR_VAL,
-		.off1	= td_var_offset(zone_size),
-	},
-	{
-		.name	= "zoneskip",
-		.type	= FIO_OPT_STR_VAL,
-		.off1	= td_var_offset(zone_skip),
-	},
-	{
-		.name	= "lockmem",
-		.type	= FIO_OPT_STR_VAL,
-		.cb	= str_lockmem_cb,
+		.help	= "Block size unit",
+		.def	= "4k",
 	},
 	{
 		.name	= "bsrange",
@@ -173,202 +145,389 @@ static struct fio_option options[] = {
 		.off2	= td_var_offset(max_bs[DDIR_READ]),
 		.off3	= td_var_offset(min_bs[DDIR_WRITE]),
 		.off4	= td_var_offset(max_bs[DDIR_WRITE]),
+		.help	= "Set block size range (in more detail than bs)",
+	},
+	{
+		.name	= "bs_unaligned",
+		.type	= FIO_OPT_STR_SET,
+		.off1	= td_var_offset(bs_unaligned),
+		.help	= "Don't sector align IO buffer sizes",
+	},
+	{
+		.name	= "offset",
+		.type	= FIO_OPT_STR_VAL,
+		.off1	= td_var_offset(start_offset),
+		.help	= "Start IO from this offset",
+		.def	= "0",
+	},
+	{
+		.name	= "randrepeat",
+		.type	= FIO_OPT_BOOL,
+		.off1	= td_var_offset(rand_repeatable),
+		.help	= "Use repeatable random IO pattern",
+		.def	= "1",
+	},
+	{
+		.name	= "norandommap",
+		.type	= FIO_OPT_STR_SET,
+		.off1	= td_var_offset(norandommap),
+		.help	= "Accept potential duplicate random blocks",
 	},
 	{
 		.name	= "nrfiles",
 		.type	= FIO_OPT_INT,
 		.off1	= td_var_offset(nr_files),
+		.help	= "Split job workload between this number of files",
+		.def	= "1",
 	},
 	{
-		.name	= "iodepth",
-		.type	= FIO_OPT_INT,
-		.off1	= td_var_offset(iodepth),
+		.name	= "file_service_type",
+		.type	= FIO_OPT_STR,
+		.off1	= td_var_offset(file_service_type),
+		.help	= "How to select which file to service next",
+		.def	= "roundrobin",
+		.posval	= {
+			  { .ival = "random", .oval = FIO_FSERVICE_RANDOM },
+			  { .ival = "roundrobin", .oval = FIO_FSERVICE_RR },
+			  },
 	},
 	{
 		.name	= "fsync",
 		.type	= FIO_OPT_INT,
 		.off1	= td_var_offset(fsync_blocks),
+		.help	= "Issue fsync for writes every given number of blocks",
+		.def	= "0",
+	},
+	{
+		.name	= "direct",
+		.type	= FIO_OPT_BOOL,
+		.off1	= td_var_offset(odirect),
+		.help	= "Use O_DIRECT IO (negates buffered)",
+		.def	= "0",
+	},
+	{
+		.name	= "buffered",
+		.type	= FIO_OPT_BOOL,
+		.off1	= td_var_offset(odirect),
+		.neg	= 1,
+		.help	= "Use buffered IO (negates direct)",
+		.def	= "1",
+	},
+	{
+		.name	= "overwrite",
+		.type	= FIO_OPT_BOOL,
+		.off1	= td_var_offset(overwrite),
+		.help	= "When writing, set whether to overwrite current data",
+		.def	= "0",
+	},
+	{
+		.name	= "loops",
+		.type	= FIO_OPT_INT,
+		.off1	= td_var_offset(loops),
+		.help	= "Number of times to run the job",
+		.def	= "1",
+	},
+	{
+		.name	= "numjobs",
+		.type	= FIO_OPT_INT,
+		.off1	= td_var_offset(numjobs),
+		.help	= "Duplicate this job this many times",
+		.def	= "1",
+	},
+	{
+		.name	= "startdelay",
+		.type	= FIO_OPT_INT,
+		.off1	= td_var_offset(start_delay),
+		.help	= "Only start job when this period has passed",
+		.def	= "0",
+	},
+	{
+		.name	= "runtime",
+		.alias	= "timeout",
+		.type	= FIO_OPT_STR_VAL_TIME,
+		.off1	= td_var_offset(timeout),
+		.help	= "Stop workload when this amount of time has passed",
+		.def	= "0",
+	},
+	{
+		.name	= "mem",
+		.type	= FIO_OPT_STR,
+		.cb	= str_mem_cb,
+		.off1	= td_var_offset(mem_type),
+		.help	= "Backing type for IO buffers",
+		.def	= "malloc",
+		.posval	= {
+			  { .ival = "malloc", .oval = MEM_MALLOC },
+			  { .ival = "shm", .oval = MEM_SHM },
+#ifdef FIO_HAVE_HUGETLB
+			  { .ival = "shmhuge", .oval = MEM_SHMHUGE },
+#endif
+			  { .ival = "mmap", .oval = MEM_MMAP },
+#ifdef FIO_HAVE_HUGETLB
+			  { .ival = "mmaphuge", .oval = MEM_MMAPHUGE },
+#endif
+			  },
+	},
+	{
+		.name	= "verify",
+		.type	= FIO_OPT_STR,
+		.off1	= td_var_offset(verify),
+		.help	= "Verify sum function",
+		.def	= "0",
+		.posval = {
+			  { .ival = "0", .oval = VERIFY_NONE },
+			  { .ival = "crc32", .oval = VERIFY_CRC32 },
+			  { .ival = "md5", .oval = VERIFY_MD5 },
+			  },
+	},
+	{
+		.name	= "write_iolog",
+		.type	= FIO_OPT_STR_STORE,
+		.off1	= td_var_offset(write_iolog_file),
+		.help	= "Store IO pattern to file",
+	},
+	{
+		.name	= "read_iolog",
+		.type	= FIO_OPT_STR_STORE,
+		.off1	= td_var_offset(read_iolog_file),
+		.help	= "Playback IO pattern from file",
+	},
+	{
+		.name	= "exec_prerun",
+		.type	= FIO_OPT_STR_STORE,
+		.off1	= td_var_offset(exec_prerun),
+		.help	= "Execute this file prior to running job",
+	},
+	{
+		.name	= "exec_postrun",
+		.type	= FIO_OPT_STR_STORE,
+		.off1	= td_var_offset(exec_postrun),
+		.help	= "Execute this file after running job",
+	},
+#ifdef FIO_HAVE_IOSCHED_SWITCH
+	{
+		.name	= "ioscheduler",
+		.type	= FIO_OPT_STR_STORE,
+		.off1	= td_var_offset(ioscheduler),
+		.help	= "Use this IO scheduler on the backing device",
+	},
+#endif
+	{
+		.name	= "zonesize",
+		.type	= FIO_OPT_STR_VAL,
+		.off1	= td_var_offset(zone_size),
+		.help	= "Give size of an IO zone",
+		.def	= "0",
+	},
+	{
+		.name	= "zoneskip",
+		.type	= FIO_OPT_STR_VAL,
+		.off1	= td_var_offset(zone_skip),
+		.help	= "Space between IO zones",
+		.def	= "0",
+	},
+	{
+		.name	= "lockmem",
+		.type	= FIO_OPT_STR_VAL,
+		.cb	= str_lockmem_cb,
+		.help	= "Lock down this amount of memory",
+		.def	= "0",
 	},
 	{
 		.name	= "rwmixcycle",
 		.type	= FIO_OPT_INT,
 		.off1	= td_var_offset(rwmixcycle),
+		.help	= "Cycle period for mixed read/write workloads (msec)",
+		.def	= "500",
 	},
 	{
 		.name	= "rwmixread",
 		.type	= FIO_OPT_INT,
 		.off1	= td_var_offset(rwmixread),
-		.max_val= 100,
+		.maxval	= 100,
+		.help	= "Percentage of mixed workload that is reads",
+		.def	= "50",
 	},
 	{
 		.name	= "rwmixwrite",
 		.type	= FIO_OPT_INT,
 		.off1	= td_var_offset(rwmixwrite),
-		.max_val= 100,
+		.maxval	= 100,
+		.help	= "Percentage of mixed workload that is writes",
+		.def	= "50",
 	},
 	{
 		.name	= "nice",
 		.type	= FIO_OPT_INT,
 		.off1	= td_var_offset(nice),
+		.help	= "Set job CPU nice value",
+		.minval	= -19,
+		.maxval	= 20,
+		.def	= "0",
 	},
 #ifdef FIO_HAVE_IOPRIO
 	{
 		.name	= "prio",
 		.type	= FIO_OPT_INT,
 		.cb	= str_prio_cb,
+		.help	= "Set job IO priority value",
+		.minval	= 0,
+		.maxval	= 7,
 	},
 	{
 		.name	= "prioclass",
 		.type	= FIO_OPT_INT,
 		.cb	= str_prioclass_cb,
+		.help	= "Set job IO priority class",
+		.minval	= 0,
+		.maxval	= 3,
 	},
 #endif
 	{
 		.name	= "thinktime",
 		.type	= FIO_OPT_INT,
 		.off1	= td_var_offset(thinktime),
+		.help	= "Idle time between IO buffers (usec)",
+		.def	= "0",
+	},
+	{
+		.name	= "thinktime_spin",
+		.type	= FIO_OPT_INT,
+		.off1	= td_var_offset(thinktime_spin),
+		.help	= "Start thinktime by spinning this amount (usec)",
+		.def	= "0",
 	},
 	{
 		.name	= "thinktime_blocks",
 		.type	= FIO_OPT_INT,
 		.off1	= td_var_offset(thinktime_blocks),
+		.help	= "IO buffer period between 'thinktime'",
+		.def	= "1",
 	},
 	{
 		.name	= "rate",
 		.type	= FIO_OPT_INT,
 		.off1	= td_var_offset(rate),
+		.help	= "Set bandwidth rate",
 	},
 	{
 		.name	= "ratemin",
 		.type	= FIO_OPT_INT,
 		.off1	= td_var_offset(ratemin),
+		.help	= "The bottom limit accepted",
 	},
 	{
 		.name	= "ratecycle",
 		.type	= FIO_OPT_INT,
 		.off1	= td_var_offset(ratecycle),
-	},
-	{
-		.name	= "startdelay",
-		.type	= FIO_OPT_INT,
-		.off1	= td_var_offset(start_delay),
-	},
-	{
-		.name	= "timeout",
-		.type	= FIO_OPT_STR_VAL_TIME,
-		.off1	= td_var_offset(timeout),
+		.help	= "Window average for rate limits (msec)",
+		.def	= "1000",
 	},
 	{
 		.name	= "invalidate",
-		.type	= FIO_OPT_INT,
+		.type	= FIO_OPT_BOOL,
 		.off1	= td_var_offset(invalidate_cache),
+		.help	= "Invalidate buffer/page cache prior to running job",
+		.def	= "1",
 	},
 	{
 		.name	= "sync",
-		.type	= FIO_OPT_INT,
+		.type	= FIO_OPT_BOOL,
 		.off1	= td_var_offset(sync_io),
+		.help	= "Use O_SYNC for buffered writes",
+		.def	= "0",
 	},
 	{
 		.name	= "bwavgtime",
 		.type	= FIO_OPT_INT,
 		.off1	= td_var_offset(bw_avg_time),
+		.help	= "Time window over which to calculate bandwidth (msec)",
+		.def	= "500",
 	},
 	{
 		.name	= "create_serialize",
-		.type	= FIO_OPT_INT,
+		.type	= FIO_OPT_BOOL,
 		.off1	= td_var_offset(create_serialize),
+		.help	= "Serialize creating of job files",
+		.def	= "1",
 	},
 	{
 		.name	= "create_fsync",
-		.type	= FIO_OPT_INT,
+		.type	= FIO_OPT_BOOL,
 		.off1	= td_var_offset(create_fsync),
-	},
-	{
-		.name	= "loops",
-		.type	= FIO_OPT_INT,
-		.off1	= td_var_offset(loops),
-	},
-	{
-		.name	= "numjobs",
-		.type	= FIO_OPT_INT,
-		.off1	= td_var_offset(numjobs),
+		.help	= "Fsync file after creation",
+		.def	= "1",
 	},
 	{
 		.name	= "cpuload",
 		.type	= FIO_OPT_INT,
 		.off1	= td_var_offset(cpuload),
+		.help	= "Use this percentage of CPU",
 	},
 	{
 		.name	= "cpuchunks",
 		.type	= FIO_OPT_INT,
 		.off1	= td_var_offset(cpucycle),
-	},
-	{
-		.name	= "direct",
-		.type	= FIO_OPT_INT,
-		.off1	= td_var_offset(odirect),
-	},
-	{
-		.name	= "overwrite",
-		.type	= FIO_OPT_INT,
-		.off1	= td_var_offset(overwrite),
+		.help	= "Length of the CPU burn cycles",
 	},
 #ifdef FIO_HAVE_CPU_AFFINITY
 	{
 		.name	= "cpumask",
 		.type	= FIO_OPT_INT,
 		.cb	= str_cpumask_cb,
+		.help	= "CPU affinity mask",
 	},
 #endif
 	{
 		.name	= "end_fsync",
-		.type	= FIO_OPT_INT,
+		.type	= FIO_OPT_BOOL,
 		.off1	= td_var_offset(end_fsync),
+		.help	= "Include fsync at the end of job",
+		.def	= "0",
 	},
 	{
 		.name	= "unlink",
-		.type	= FIO_OPT_INT,
+		.type	= FIO_OPT_BOOL,
 		.off1	= td_var_offset(unlink),
+		.help	= "Unlink created files after job has completed",
+		.def	= "0",
 	},
 	{
 		.name	= "exitall",
 		.type	= FIO_OPT_STR_SET,
 		.cb	= str_exitall_cb,
+		.help	= "Terminate all jobs when one exits",
 	},
 	{
 		.name	= "stonewall",
 		.type	= FIO_OPT_STR_SET,
 		.off1	= td_var_offset(stonewall),
+		.help	= "Insert a hard barrier between this job and previous",
 	},
 	{
 		.name	= "thread",
 		.type	= FIO_OPT_STR_SET,
-		.off1	= td_var_offset(thread),
+		.off1	= td_var_offset(use_thread),
+		.help	= "Use threads instead of forks",
 	},
 	{
 		.name	= "write_bw_log",
 		.type	= FIO_OPT_STR_SET,
 		.off1	= td_var_offset(write_bw_log),
+		.help	= "Write log of bandwidth during run",
 	},
 	{
 		.name	= "write_lat_log",
 		.type	= FIO_OPT_STR_SET,
 		.off1	= td_var_offset(write_lat_log),
-	},
-	{
-		.name	= "norandommap",
-		.type	= FIO_OPT_STR_SET,
-		.off1	= td_var_offset(norandommap),
-	},
-	{
-		.name	= "bs_unaligned",
-		.type	= FIO_OPT_STR_SET,
-		.off1	= td_var_offset(bs_unaligned),
+		.help	= "Write log of latency during run",
 	},
 	{
 		.name	= "hugepage-size",
 		.type	= FIO_OPT_STR_VAL,
 		.off1	= td_var_offset(hugepage_size),
+		.help	= "When using hugepages, specify size of each page",
+		.def	= __stringify(FIO_HUGE_PAGE),
 	},
 	{
 		.name = NULL,
@@ -414,14 +573,24 @@ static struct option long_options[FIO_JOB_OPTS + FIO_CMD_OPTS] = {
 		.has_arg	= no_argument,
 		.val		= 'v',
 	},
+	{
+		.name		= "help",
+		.has_arg	= no_argument,
+		.val		= 'h',
+	},
+	{
+		.name		= "cmdhelp",
+		.has_arg	= required_argument,
+		.val		= 'c',
+	},
 	{
 		.name		= NULL,
 	},
 };
 
-static int def_timeout = DEF_TIMEOUT;
+static int def_timeout = 0;
 
-static char fio_version_string[] = "fio 1.11";
+static char fio_version_string[] = "fio 1.12";
 
 static char **ini_file;
 static int max_jobs = MAX_JOBS;
@@ -435,8 +604,8 @@ unsigned long long mlock_size = 0;
 FILE *f_out = NULL;
 FILE *f_err = NULL;
 
-static int write_lat_log = DEF_WRITE_LAT_LOG;
-static int write_bw_log = DEF_WRITE_BW_LOG;
+static int write_lat_log = 0;
+int write_bw_log = 0;
 
 /*
  * Return a free job structure.
@@ -462,6 +631,9 @@ static void put_job(struct thread_data *td)
 	if (td == &def_thread)
 		return;
 
+	if (td->error)
+		fprintf(f_out, "fio: %s\n", td->verror);
+
 	memset(&threads[td->thread_number - 1], 0, sizeof(*td));
 	thread_number--;
 }
@@ -491,7 +663,7 @@ static void fixup_options(struct thread_data *td)
 	/*
 	 * only really works for sequential io for now, and with 1 file
 	 */
-	if (td->zone_size && !td->sequential && td->nr_files == 1)
+	if (td->zone_size && td_random(td) && td->nr_files == 1)
 		td->zone_size = 0;
 
 	/*
@@ -526,6 +698,33 @@ static void fixup_options(struct thread_data *td)
 	 */
 	if (td->filetype == FIO_TYPE_CHAR && td->odirect)
 		td->odirect = 0;
+
+	/*
+	 * thinktime_spin must be less than thinktime
+	 */
+	if (td->thinktime_spin > td->thinktime)
+		td->thinktime_spin = td->thinktime;
+
+	/*
+	 * The low water mark cannot be bigger than the iodepth
+	 */
+	if (td->iodepth_low > td->iodepth || !td->iodepth_low) {
+		/*
+		 * syslet work around - if the workload is sequential,
+		 * we want to let the queue drain all the way down to
+		 * avoid seeking between async threads
+		 */
+		if (!strcmp(td->io_ops->name, "syslet-rw") && !td_random(td))
+			td->iodepth_low = 1;
+		else
+			td->iodepth_low = td->iodepth;
+	}
+
+	/*
+	 * If batch number isn't set, default to the same as iodepth
+	 */
+	if (td->iodepth_batch > td->iodepth || !td->iodepth_batch)
+		td->iodepth_batch = td->iodepth;
 }
 
 /*
@@ -534,7 +733,7 @@ static void fixup_options(struct thread_data *td)
 static char *to_kmg(unsigned int val)
 {
 	char *buf = malloc(32);
-	char post[] = { 0, 'K', 'M', 'G', 'P', 0 };
+	char post[] = { 0, 'K', 'M', 'G', 'P', 'E', 0 };
 	char *p = post;
 
 	do {
@@ -556,10 +755,10 @@ static char *to_kmg(unsigned int val)
  */
 static int add_job(struct thread_data *td, const char *jobname, int job_add_num)
 {
-	const char *ddir_str[] = { "read", "write", "randread", "randwrite",
-				   "rw", NULL, "randrw" };
+	const char *ddir_str[] = { NULL, "read", "write", "rw", NULL,
+				   "randread", "randwrite", "randrw" };
 	struct stat sb;
-	int numjobs, ddir, i;
+	int numjobs, i;
 	struct fio_file *f;
 
 	/*
@@ -568,22 +767,13 @@ static int add_job(struct thread_data *td, const char *jobname, int job_add_num)
 	if (td == &def_thread)
 		return 0;
 
-	/*
-	 * Set default io engine, if none set
-	 */
-	if (!td->io_ops) {
-		td->io_ops = load_ioengine(td, DEF_IO_ENGINE_NAME);
-		if (!td->io_ops) {
-			log_err("default engine %s not there?\n", DEF_IO_ENGINE_NAME);
-			return 1;
-		}
-	}
+	assert(td->io_ops);
 
 	if (td->odirect)
 		td->io_ops->flags |= FIO_RAWIO;
 
 	td->filetype = FIO_TYPE_FILE;
-	if (!stat(jobname, &sb)) {
+	if (td->filename && !lstat(td->filename, &sb)) {
 		if (S_ISBLK(sb.st_mode))
 			td->filetype = FIO_TYPE_BD;
 		else if (S_ISCHR(sb.st_mode))
@@ -601,8 +791,18 @@ static int add_job(struct thread_data *td, const char *jobname, int job_add_num)
 		char tmp[PATH_MAX];
 		int len = 0;
 
-		if (td->directory && td->directory[0] != '\0')
+		if (td->directory && td->directory[0] != '\0') {
+			if (lstat(td->directory, &sb) < 0) {
+				log_err("fio: %s is not a directory\n", td->directory);
+				td_verror(td, errno, "lstat");
+				return 1;
+			}
+			if (!S_ISDIR(sb.st_mode)) {
+				log_err("fio: %s is not a directory\n", td->directory);
+				return 1;
+			}
 			len = sprintf(tmp, "%s/", td->directory);
+		}
 
 		td->files = malloc(sizeof(struct fio_file) * td->nr_files);
 
@@ -633,9 +833,9 @@ static int add_job(struct thread_data *td, const char *jobname, int job_add_num)
 		
 	fio_sem_init(&td->mutex, 0);
 
-	td->clat_stat[0].min_val = td->clat_stat[1].min_val = ULONG_MAX;
-	td->slat_stat[0].min_val = td->slat_stat[1].min_val = ULONG_MAX;
-	td->bw_stat[0].min_val = td->bw_stat[1].min_val = ULONG_MAX;
+	td->ts.clat_stat[0].min_val = td->ts.clat_stat[1].min_val = ULONG_MAX;
+	td->ts.slat_stat[0].min_val = td->ts.slat_stat[1].min_val = ULONG_MAX;
+	td->ts.bw_stat[0].min_val = td->ts.bw_stat[1].min_val = ULONG_MAX;
 
 	if (td->stonewall && td->thread_number > 1)
 		groupid++;
@@ -646,17 +846,15 @@ static int add_job(struct thread_data *td, const char *jobname, int job_add_num)
 		goto err;
 
 	if (td->write_lat_log) {
-		setup_log(&td->slat_log);
-		setup_log(&td->clat_log);
+		setup_log(&td->ts.slat_log);
+		setup_log(&td->ts.clat_log);
 	}
 	if (td->write_bw_log)
-		setup_log(&td->bw_log);
+		setup_log(&td->ts.bw_log);
 
 	if (!td->name)
 		td->name = strdup(jobname);
 
-	ddir = td->ddir + (!td->sequential << 1) + (td->iomix << 2);
-
 	if (!terse_output) {
 		if (!job_add_num) {
 			if (td->io_ops->flags & FIO_CPUIO)
@@ -669,7 +867,7 @@ static int add_job(struct thread_data *td, const char *jobname, int job_add_num)
 				c3 = to_kmg(td->min_bs[DDIR_WRITE]);
 				c4 = to_kmg(td->max_bs[DDIR_WRITE]);
 
-				fprintf(f_out, "%s: (g=%d): rw=%s, odir=%u, bs=%s-%s/%s-%s, rate=%u, ioengine=%s, iodepth=%u\n", td->name, td->groupid, ddir_str[ddir], td->odirect, c1, c2, c3, c4, td->rate, td->io_ops->name, td->iodepth);
+				fprintf(f_out, "%s: (g=%d): rw=%s, bs=%s-%s/%s-%s, ioengine=%s, iodepth=%u\n", td->name, td->groupid, ddir_str[td->td_ddir], c1, c2, c3, c4, td->io_ops->name, td->iodepth);
 
 				free(c1);
 				free(c2);
@@ -710,7 +908,7 @@ err:
  */
 int init_random_state(struct thread_data *td)
 {
-	unsigned long seeds[4];
+	unsigned long seeds[5];
 	int fd, num_maps, blocks, i;
 	struct fio_file *f;
 
@@ -719,12 +917,12 @@ int init_random_state(struct thread_data *td)
 
 	fd = open("/dev/urandom", O_RDONLY);
 	if (fd == -1) {
-		td_verror(td, errno);
+		td_verror(td, errno, "open");
 		return 1;
 	}
 
 	if (read(fd, seeds, sizeof(seeds)) < (int) sizeof(seeds)) {
-		td_verror(td, EIO);
+		td_verror(td, EIO, "read");
 		close(fd);
 		return 1;
 	}
@@ -735,15 +933,18 @@ int init_random_state(struct thread_data *td)
 	os_random_seed(seeds[1], &td->verify_state);
 	os_random_seed(seeds[2], &td->rwmix_state);
 
-	if (td->sequential)
+	if (td->file_service_type == FIO_FSERVICE_RANDOM)
+		os_random_seed(seeds[3], &td->next_file_state);
+
+	if (!td_random(td))
 		return 0;
 
 	if (td->rand_repeatable)
-		seeds[3] = DEF_RANDSEED;
+		seeds[4] = FIO_RANDSEED * td->thread_number;
 
 	if (!td->norandommap) {
 		for_each_file(td, f, i) {
-			blocks = (f->file_size + td->rw_min_bs - 1) / td->rw_min_bs;
+			blocks = (f->real_file_size + td->rw_min_bs - 1) / td->rw_min_bs;
 			num_maps = (blocks + BLOCKS_PER_MAP-1)/ BLOCKS_PER_MAP;
 			f->file_map = malloc(num_maps * sizeof(long));
 			f->num_maps = num_maps;
@@ -751,7 +952,7 @@ int init_random_state(struct thread_data *td)
 		}
 	}
 
-	os_random_seed(seeds[3], &td->random_state);
+	os_random_seed(seeds[4], &td->random_state);
 	return 0;
 }
 
@@ -776,6 +977,8 @@ static int is_empty_or_comment(char *line)
 	for (i = 0; i < strlen(line); i++) {
 		if (line[i] == ';')
 			return 1;
+		if (line[i] == '#')
+			return 1;
 		if (!isspace(line[i]) && !iscntrl(line[i]))
 			return 0;
 	}
@@ -783,61 +986,6 @@ static int is_empty_or_comment(char *line)
 	return 1;
 }
 
-static int str_rw_cb(void *data, const char *mem)
-{
-	struct thread_data *td = data;
-
-	if (!strncmp(mem, "read", 4) || !strncmp(mem, "0", 1)) {
-		td->ddir = DDIR_READ;
-		td->sequential = 1;
-		return 0;
-	} else if (!strncmp(mem, "randread", 8)) {
-		td->ddir = DDIR_READ;
-		td->sequential = 0;
-		return 0;
-	} else if (!strncmp(mem, "write", 5) || !strncmp(mem, "1", 1)) {
-		td->ddir = DDIR_WRITE;
-		td->sequential = 1;
-		return 0;
-	} else if (!strncmp(mem, "randwrite", 9)) {
-		td->ddir = DDIR_WRITE;
-		td->sequential = 0;
-		return 0;
-	} else if (!strncmp(mem, "rw", 2)) {
-		td->ddir = DDIR_READ;
-		td->iomix = 1;
-		td->sequential = 1;
-		return 0;
-	} else if (!strncmp(mem, "randrw", 6)) {
-		td->ddir = DDIR_READ;
-		td->iomix = 1;
-		td->sequential = 0;
-		return 0;
-	}
-
-	log_err("fio: data direction: read, write, randread, randwrite, rw, randrw\n");
-	return 1;
-}
-
-static int str_verify_cb(void *data, const char *mem)
-{
-	struct thread_data *td = data;
-
-	if (!strncmp(mem, "0", 1)) {
-		td->verify = VERIFY_NONE;
-		return 0;
-	} else if (!strncmp(mem, "md5", 3) || !strncmp(mem, "1", 1)) {
-		td->verify = VERIFY_MD5;
-		return 0;
-	} else if (!strncmp(mem, "crc32", 5)) {
-		td->verify = VERIFY_CRC32;
-		return 0;
-	}
-
-	log_err("fio: verify types: md5, crc32\n");
-	return 1;
-}
-
 /*
  * Check if mmap/mmaphuge has a :/foo/bar/file at the end. If so, return that.
  */
@@ -858,49 +1006,15 @@ static int str_mem_cb(void *data, const char *mem)
 {
 	struct thread_data *td = data;
 
-	if (!strncmp(mem, "malloc", 6)) {
-		td->mem_type = MEM_MALLOC;
-		return 0;
-	} else if (!strncmp(mem, "mmaphuge", 8)) {
-#ifdef FIO_HAVE_HUGETLB
-		/*
-		 * mmaphuge must be appended with the actual file
-		 */
+	if (td->mem_type == MEM_MMAPHUGE || td->mem_type == MEM_MMAP) {
 		td->mmapfile = get_mmap_file(mem);
-		if (!td->mmapfile) {
+		if (td->mem_type == MEM_MMAPHUGE && !td->mmapfile) {
 			log_err("fio: mmaphuge:/path/to/file\n");
 			return 1;
 		}
-
-		td->mem_type = MEM_MMAPHUGE;
-		return 0;
-#else
-		log_err("fio: mmaphuge not available\n");
-		return 1;
-#endif
-	} else if (!strncmp(mem, "mmap", 4)) {
-		/*
-		 * Check if the user wants file backed memory. It's ok
-		 * if there's no file given, we'll just use anon mamp then.
-		 */
-		td->mmapfile = get_mmap_file(mem);
-		td->mem_type = MEM_MMAP;
-		return 0;
-	} else if (!strncmp(mem, "shmhuge", 7)) {
-#ifdef FIO_HAVE_HUGETLB
-		td->mem_type = MEM_SHMHUGE;
-		return 0;
-#else
-		log_err("fio: shmhuge not available\n");
-		return 1;
-#endif
-	} else if (!strncmp(mem, "shm", 3)) {
-		td->mem_type = MEM_SHM;
-		return 0;
 	}
 
-	log_err("fio: mem type: malloc, shm, shmhuge, mmap, mmaphuge\n");
-	return 1;
+	return 0;
 }
 
 static int str_ioengine_cb(void *data, const char *str)
@@ -911,8 +1025,6 @@ static int str_ioengine_cb(void *data, const char *str)
 	if (td->io_ops)
 		return 0;
 
-	log_err("fio: ioengine= libaio, posixaio, sync, mmap, sgio, splice, cpu, null\n");
-	log_err("fio: or specify path to dynamic ioengine module\n");
 	return 1;
 }
 
@@ -1052,41 +1164,14 @@ static int fill_def_thread(void)
 	}
 
 	/*
-	 * fill globals
+	 * fill default options
 	 */
-	def_thread.ddir = DDIR_READ;
-	def_thread.iomix = 0;
-	def_thread.bs[DDIR_READ] = DEF_BS;
-	def_thread.bs[DDIR_WRITE] = DEF_BS;
-	def_thread.min_bs[DDIR_READ] = def_thread.min_bs[DDIR_WRITE] = 0;
-	def_thread.max_bs[DDIR_READ] = def_thread.max_bs[DDIR_WRITE] = 0;
-	def_thread.odirect = DEF_ODIRECT;
-	def_thread.ratecycle = DEF_RATE_CYCLE;
-	def_thread.sequential = DEF_SEQUENTIAL;
+	fill_default_options(&def_thread, options);
+
 	def_thread.timeout = def_timeout;
-	def_thread.overwrite = DEF_OVERWRITE;
-	def_thread.invalidate_cache = DEF_INVALIDATE;
-	def_thread.sync_io = DEF_SYNCIO;
-	def_thread.mem_type = MEM_MALLOC;
-	def_thread.bw_avg_time = DEF_BWAVGTIME;
-	def_thread.create_serialize = DEF_CREATE_SER;
-	def_thread.create_fsync = DEF_CREATE_FSYNC;
-	def_thread.loops = DEF_LOOPS;
-	def_thread.verify = DEF_VERIFY;
-	def_thread.stonewall = DEF_STONEWALL;
-	def_thread.numjobs = DEF_NUMJOBS;
-	def_thread.use_thread = DEF_USE_THREAD;
-	def_thread.rwmixcycle = DEF_RWMIX_CYCLE;
-	def_thread.rwmixread = DEF_RWMIX_READ;
-	def_thread.nice = DEF_NICE;
-	def_thread.rand_repeatable = DEF_RAND_REPEAT;
-	def_thread.nr_files = DEF_NR_FILES;
-	def_thread.unlink = DEF_UNLINK;
 	def_thread.write_bw_log = write_bw_log;
 	def_thread.write_lat_log = write_lat_log;
-	def_thread.norandommap = DEF_NO_RAND_MAP;
-	def_thread.hugepage_size = DEF_HUGEPAGE_SIZE;
-	def_thread.thinktime_blocks = DEF_THINKTIME_BLOCKS;
+
 #ifdef FIO_HAVE_DISK_UTIL
 	def_thread.do_disk_util = 1;
 #endif
@@ -1103,12 +1188,14 @@ static void usage(void)
 	printf("\t--bandwidth-log\tGenerate per-job bandwidth logs\n");
 	printf("\t--minimal\tMinimal (terse) output\n");
 	printf("\t--version\tPrint version info and exit\n");
+	printf("\t--help\t\tPrint this page\n");
+	printf("\t--cmdhelp=cmd\tPrint command help, \"all\" for all of them\n");
 }
 
 static int parse_cmd_line(int argc, char *argv[])
 {
 	struct thread_data *td = NULL;
-	int c, ini_idx = 0, lidx, ret;
+	int c, ini_idx = 0, lidx, ret, dont_add_job = 0;
 
 	while ((c = getopt_long(argc, argv, "", long_options, &lidx)) != -1) {
 		switch (c) {
@@ -1135,6 +1222,9 @@ static int parse_cmd_line(int argc, char *argv[])
 		case 'h':
 			usage();
 			exit(0);
+		case 'c':
+			ret = show_cmd_help(options, optarg);
+			exit(ret);
 		case 'v':
 			printf("%s\n", fio_version_string);
 			exit(0);
@@ -1159,11 +1249,8 @@ static int parse_cmd_line(int argc, char *argv[])
 			}
 
 			ret = parse_cmd_option(opt, val, options, td);
-			if (ret) {
-				log_err("fio: job dropped\n");
-				put_job(td);
-				td = NULL;
-			}
+			if (ret)
+				dont_add_job = 1;
 			break;
 		}
 		default:
@@ -1172,9 +1259,13 @@ static int parse_cmd_line(int argc, char *argv[])
 	}
 
 	if (td) {
-		ret = add_job(td, td->name ?: "fio", 0);
-		if (ret)
+		if (dont_add_job)
 			put_job(td);
+		else {
+			ret = add_job(td, td->name ?: "fio", 0);
+			if (ret)
+				put_job(td);
+		}
 	}
 
 	while (optind < argc) {
@@ -1271,6 +1362,8 @@ int parse_options(int argc, char *argv[])
 	f_out = stdout;
 	f_err = stderr;
 
+	options_init(options);
+
 	dupe_job_options();
 
 	if (setup_thread_area())
