@@ -14,6 +14,7 @@
 #include <unistd.h>
 #include <string.h>
 #include <dlfcn.h>
+#include <assert.h>
 
 #include "fio.h"
 #include "os.h"
@@ -33,16 +34,27 @@ static int check_engine_ops(struct ioengine_ops *ops)
 	if (ops->flags & FIO_CPUIO)
 		return 0;
 
+	if (!ops->queue) {
+		log_err("%s: no queue handler\n", ops->name);
+		return 1;
+	}
+
+	/*
+	 * sync engines only need a ->queue()
+	 */
+	if (ops->flags & FIO_SYNCIO)
+		return 0;
+	
 	if (!ops->event) {
-		log_err("%s: no event handler)\n", ops->name);
+		log_err("%s: no event handler\n", ops->name);
 		return 1;
 	}
 	if (!ops->getevents) {
-		log_err("%s: no getevents handler)\n", ops->name);
+		log_err("%s: no getevents handler\n", ops->name);
 		return 1;
 	}
 	if (!ops->queue) {
-		log_err("%s: no queue handler)\n", ops->name);
+		log_err("%s: no queue handler\n", ops->name);
 		return 1;
 	}
 		
@@ -55,14 +67,10 @@ void unregister_ioengine(struct ioengine_ops *ops)
 	INIT_LIST_HEAD(&ops->list);
 }
 
-int register_ioengine(struct ioengine_ops *ops)
+void register_ioengine(struct ioengine_ops *ops)
 {
-	if (check_engine_ops(ops))
-		return 1;
-
 	INIT_LIST_HEAD(&ops->list);
 	list_add_tail(&ops->list, &engine_list);
-	return 0;
 }
 
 static struct ioengine_ops *find_ioengine(const char *name)
@@ -94,7 +102,7 @@ static struct ioengine_ops *dlopen_ioengine(struct thread_data *td,
 	dlerror();
 	dlhandle = dlopen(engine_lib, RTLD_LAZY);
 	if (!dlhandle) {
-		td_vmsg(td, -1, dlerror());
+		td_vmsg(td, -1, dlerror(), "dlopen");
 		return NULL;
 	}
 
@@ -104,7 +112,7 @@ static struct ioengine_ops *dlopen_ioengine(struct thread_data *td,
 	 */
 	ops = dlsym(dlhandle, "ioengine");
 	if (!ops) {
-		td_vmsg(td, -1, dlerror());
+		td_vmsg(td, -1, dlerror(), "dlsym");
 		dlclose(dlhandle);
 		return NULL;
 	}
@@ -156,14 +164,17 @@ void close_ioengine(struct thread_data *td)
 	if (td->io_ops->dlhandle)
 		dlclose(td->io_ops->dlhandle);
 
+#if 0
+	/* we can't do this for threads, so just leak it, it's exiting */
 	free(td->io_ops);
+#endif
 	td->io_ops = NULL;
 }
 
 int td_io_prep(struct thread_data *td, struct io_u *io_u)
 {
-	if (td->io_ops->prep && td->io_ops->prep(td, io_u))
-		return 1;
+	if (td->io_ops->prep)
+		return td->io_ops->prep(td, io_u);
 
 	return 0;
 }
@@ -171,14 +182,63 @@ int td_io_prep(struct thread_data *td, struct io_u *io_u)
 int td_io_getevents(struct thread_data *td, int min, int max,
 		    struct timespec *t)
 {
-	return td->io_ops->getevents(td, min, max, t);
+	if (min > 0 && td->io_ops->commit) {
+		int r = td->io_ops->commit(td);
+
+		if (r < 0)
+			return r;
+	}
+	if (td->io_ops->getevents)
+		return td->io_ops->getevents(td, min, max, t);
+
+	return 0;
 }
 
 int td_io_queue(struct thread_data *td, struct io_u *io_u)
 {
-	fio_gettime(&io_u->issue_time, NULL);
+	int ret;
 
-	return td->io_ops->queue(td, io_u);
+	assert((io_u->flags & IO_U_F_FLIGHT) == 0);
+	io_u->flags |= IO_U_F_FLIGHT;
+
+	if (td->io_ops->flags & FIO_SYNCIO) {
+		fio_gettime(&io_u->issue_time, NULL);
+
+		/*
+		 * for a sync engine, set the timeout upfront
+		 */
+		if (mtime_since(&td->timeout_end, &io_u->issue_time) < IO_U_TIMEOUT)
+			io_u_set_timeout(td);
+	}
+
+	if (io_u->ddir != DDIR_SYNC)
+		td->io_issues[io_u->ddir]++;
+
+	ret = td->io_ops->queue(td, io_u);
+
+	if (ret == FIO_Q_QUEUED) {
+		int r;
+
+		td->io_u_queued++;
+		if (td->io_u_queued > td->iodepth_batch) {
+			r = td_io_commit(td);
+			if (r < 0)
+				return r;
+		}
+	}
+
+	if ((td->io_ops->flags & FIO_SYNCIO) == 0) {
+		fio_gettime(&io_u->issue_time, NULL);
+
+		/*
+		 * async engine, set the timeout here
+		 */
+		if (ret == FIO_Q_QUEUED &&
+		    mtime_since(&td->timeout_end, &io_u->issue_time) < IO_U_TIMEOUT)
+			io_u_set_timeout(td);
+	}
+
+	return ret;
 }
 
 int td_io_init(struct thread_data *td)
@@ -188,3 +248,15 @@ int td_io_init(struct thread_data *td)
 
 	return 0;
 }
+
+int td_io_commit(struct thread_data *td)
+{
+	if (!td->cur_depth)
+		return 0;
+
+	td->io_u_queued = 0;
+	if (td->io_ops->commit)
+		return td->io_ops->commit(td);
+
+	return 0;
+}
