@@ -5,7 +5,7 @@
 
 void write_iolog_put(struct thread_data *td, struct io_u *io_u)
 {
-	fprintf(td->iolog_f, "%u,%llu,%u\n", io_u->ddir, io_u->offset, io_u->buflen);
+	fprintf(td->iolog_f, "%u,%llu,%lu\n", io_u->ddir, io_u->offset, io_u->buflen);
 }
 
 int read_iolog_get(struct thread_data *td, struct io_u *io_u)
@@ -56,7 +56,7 @@ void log_io_piece(struct thread_data *td, struct io_u *io_u)
 	 * be laid out with the block scattered as written. it's faster to
 	 * read them in in that order again, so don't sort
 	 */
-	if (td->sequential || !td->overwrite) {
+	if (!td_random(td) || !td->overwrite) {
 		list_add_tail(&ipo->list, &td->io_hist_list);
 		return;
 	}
@@ -137,11 +137,11 @@ static int init_iolog_read(struct thread_data *td)
 	if (!reads && !writes)
 		return 1;
 	else if (reads && !writes)
-		td->ddir = DDIR_READ;
+		td->td_ddir = TD_DDIR_READ;
 	else if (!reads && writes)
-		td->ddir = DDIR_READ;
+		td->td_ddir = TD_DDIR_READ;
 	else
-		td->iomix = 1;
+		td->td_ddir = TD_DDIR_RW;
 
 	return 0;
 }
@@ -185,7 +185,9 @@ int init_iolog(struct thread_data *td)
 
 int setup_rate(struct thread_data *td)
 {
-	int nr_reads_per_sec;
+	unsigned long long rate;
+	int nr_reads_per_msec;
+	unsigned int bs;
 
 	if (!td->rate)
 		return 0;
@@ -195,8 +197,21 @@ int setup_rate(struct thread_data *td)
 		return -1;
 	}
 
-	nr_reads_per_sec = (td->rate * 1024) / td->min_bs[DDIR_READ];
-	td->rate_usec_cycle = 1000000 / nr_reads_per_sec;
+	if (td_rw(td))
+		bs = td->rw_min_bs;
+	else if (td_read(td))
+		bs = td->min_bs[DDIR_READ];
+	else
+		bs = td->min_bs[DDIR_WRITE];
+
+	rate = td->rate;
+	nr_reads_per_msec = (rate * 1024 * 1000) / bs;
+	if (!nr_reads_per_msec) {
+		log_err("rate lower than supported\n");
+		return -1;
+	}
+
+	td->rate_usec_cycle = 1000000000ULL / nr_reads_per_msec;
 	td->rate_pending_usleep = 0;
 	return 0;
 }
@@ -211,14 +226,12 @@ void setup_log(struct io_log **log)
 	*log = l;
 }
 
-void finish_log(struct thread_data *td, struct io_log *log, const char *name)
+void __finish_log(struct io_log *log, const char *name)
 {
-	char file_name[256];
-	FILE *f;
 	unsigned int i;
+	FILE *f;
 
-	snprintf(file_name, 200, "client%d_%s.log", td->thread_number, name);
-	f = fopen(file_name, "w");
+	f = fopen(name, "w");
 	if (!f) {
 		perror("fopen log");
 		return;
@@ -231,3 +244,11 @@ void finish_log(struct thread_data *td, struct io_log *log, const char *name)
 	free(log->log);
 	free(log);
 }
+
+void finish_log(struct thread_data *td, struct io_log *log, const char *name)
+{
+	char file_name[256];
+
+	snprintf(file_name, 200, "client%d_%s.log", td->thread_number, name);
+	__finish_log(log, file_name);
+}
