@@ -20,6 +20,13 @@
 #define FIO_HAVE_ODIRECT
 #define FIO_HAVE_HUGETLB
 
+/*
+ * Only for x86 currently
+ */
+#if defined(__i386__)
+#define FIO_HAVE_SYSLET
+#endif
+
 #define OS_MAP_ANON		(MAP_ANONYMOUS)
 
 typedef cpu_set_t os_cpu_mask_t;
@@ -73,6 +80,39 @@ static inline int vmsplice(int fd, const struct iovec *iov,
 
 #define SPLICE_DEF_SIZE	(64*1024)
 
+#ifdef FIO_HAVE_SYSLET
+
+struct syslet_uatom;
+struct async_head_user;
+
+/*
+ * syslet stuff
+ */
+static inline struct syslet_uatom *
+async_exec(struct syslet_uatom *atom, struct async_head_user *ahu)
+{
+	return (void *) syscall(__NR_async_exec, atom, ahu);
+}
+
+static inline long
+async_wait(unsigned long min_wait_events, unsigned long user_ring_idx,
+	   struct async_head_user *ahu)
+{
+	return syscall(__NR_async_wait, min_wait_events,
+			user_ring_idx, ahu);
+}
+
+static inline long async_thread(void *event, struct async_head_user *ahu)
+{
+	return syscall(__NR_async_thread, event, ahu);
+}
+
+static inline long umem_add(unsigned long *uptr, unsigned long inc)
+{
+	return syscall(__NR_umem_add, uptr, inc);
+}
+#endif /* FIO_HAVE_SYSLET */
+
 enum {
 	IOPRIO_WHO_PROCESS = 1,
 	IOPRIO_WHO_PGRP,
