@@ -11,6 +11,35 @@
 
 #include "parse.h"
 
+static void show_option_range(struct fio_option *o)
+{
+	if (!o->minval && !o->maxval)
+		return;
+
+	printf("%20s: min=%d, max=%d\n", "range", o->minval, o->maxval);
+}
+
+static void show_option_values(struct fio_option *o)
+{
+	const char *msg;
+	int i = 0;
+
+	do {
+		msg = o->posval[i].ival;
+		if (!msg)
+			break;
+
+		if (!i)
+			printf("%20s: ", "valid values");
+
+		printf("%s,", msg);
+		i++;
+	} while (i < PARSE_MAX_VP);
+
+	if (i)
+		printf("\n");
+}
+
 static unsigned long get_mult_time(char c)
 {
 	switch (c) {
@@ -40,6 +69,9 @@ static unsigned long get_mult_bytes(char c)
 		case 'g':
 		case 'G':
 			return 1024 * 1024 * 1024;
+		case 'e':
+		case 'E':
+			return 1024 * 1024 * 1024 * 1024UL;
 		default:
 			return 1;
 	}
@@ -48,7 +80,7 @@ static unsigned long get_mult_bytes(char c)
 /*
  * convert string into decimal value, noting any size suffix
  */
-static int str_to_decimal(const char *str, unsigned long long *val, int kilo)
+static int str_to_decimal(const char *str, long long *val, int kilo)
 {
 	int len;
 
@@ -56,8 +88,8 @@ static int str_to_decimal(const char *str, unsigned long long *val, int kilo)
 	if (!len)
 		return 1;
 
-	*val = strtoul(str, NULL, 10);
-	if (*val == ULONG_MAX && errno == ERANGE)
+	*val = strtoll(str, NULL, 10);
+	if (*val == LONG_MAX && errno == ERANGE)
 		return 1;
 
 	if (kilo)
@@ -68,12 +100,12 @@ static int str_to_decimal(const char *str, unsigned long long *val, int kilo)
 	return 0;
 }
 
-static int check_str_bytes(const char *p, unsigned long long *val)
+static int check_str_bytes(const char *p, long long *val)
 {
 	return str_to_decimal(p, val, 1);
 }
 
-static int check_str_time(const char *p, unsigned long long *val)
+static int check_str_time(const char *p, long long *val)
 {
 	return str_to_decimal(p, val, 0);
 }
@@ -96,7 +128,7 @@ void strip_blank_end(char *p)
 	*(s + 1) = '\0';
 }
 
-static int check_range_bytes(const char *str, unsigned long *val)
+static int check_range_bytes(const char *str, long *val)
 {
 	char suffix;
 
@@ -114,7 +146,7 @@ static int check_range_bytes(const char *str, unsigned long *val)
 	return 1;
 }
 
-static int check_int(const char *p, unsigned int *val)
+static int check_int(const char *p, int *val)
 {
 	if (!strlen(p))
 		return 1;
@@ -127,13 +159,13 @@ static int check_int(const char *p, unsigned int *val)
 static struct fio_option *find_option(struct fio_option *options,
 				      const char *opt)
 {
-	struct fio_option *o = &options[0];
+	struct fio_option *o;
 
-	while (o->name) {
+	for (o = &options[0]; o->name; o++) {
 		if (!strcmp(o->name, opt))
 			return o;
-
-		o++;
+		else if (o->alias && !strcmp(o->alias, opt))
+			return o;
 	}
 
 	return NULL;
@@ -148,9 +180,9 @@ static struct fio_option *find_option(struct fio_option *options,
 static int __handle_option(struct fio_option *o, const char *ptr, void *data,
 			   int first, int more)
 {
-	unsigned int il, *ilp;
-	unsigned long long ull, *ullp;
-	unsigned long ul1, ul2;
+	int il, *ilp;
+	long long ull, *ullp;
+	long ul1, ul2;
 	char **cp;
 	int ret = 0, is_time = 0;
 
@@ -162,8 +194,27 @@ static int __handle_option(struct fio_option *o, const char *ptr, void *data,
 	switch (o->type) {
 	case FIO_OPT_STR: {
 		fio_opt_str_fn *fn = o->cb;
+		const struct value_pair *vp;
+		int i;
 
-		ret = fn(data, ptr);
+		ret = 1;
+		for (i = 0; i < PARSE_MAX_VP; i++) {
+			vp = &o->posval[i];
+			if (!vp->ival || vp->ival[0] == '\0')
+				break;
+			if (!strncmp(vp->ival, ptr, strlen(vp->ival))) {
+				ret = 0;
+				if (!o->off1)
+					break;
+				val_store(ilp, vp->oval, o->off1, data);
+				break;
+			}
+		}
+
+		if (ret)
+			show_option_values(o);
+		else if (fn)
+			ret = fn(data, ptr);
 		break;
 	}
 	case FIO_OPT_STR_VAL_TIME:
@@ -180,8 +231,14 @@ static int __handle_option(struct fio_option *o, const char *ptr, void *data,
 		if (ret)
 			break;
 
-		if (o->max_val && ull > o->max_val)
-			ull = o->max_val;
+		if (o->maxval && ull > o->maxval) {
+			fprintf(stderr, "max value out of range: %lld (%d max)\n", ull, o->maxval);
+			return 1;
+		}
+		if (o->minval && ull < o->minval) {
+			fprintf(stderr, "min value out of range: %lld (%d min)\n", ull, o->minval);
+			return 1;
+		}
 
 		if (fn)
 			ret = fn(data, &ull);
@@ -212,8 +269,11 @@ static int __handle_option(struct fio_option *o, const char *ptr, void *data,
 
 		p1 = strchr(tmp, '-');
 		if (!p1) {
-			ret = 1;
-			break;
+			p1 = strchr(tmp, ':');
+			if (!p1) {
+				ret = 1;
+				break;
+			}
 		}
 
 		p2 = p1 + 1;
@@ -242,15 +302,25 @@ static int __handle_option(struct fio_option *o, const char *ptr, void *data,
 
 		break;
 	}
-	case FIO_OPT_INT: {
+	case FIO_OPT_INT:
+	case FIO_OPT_BOOL: {
 		fio_opt_int_fn *fn = o->cb;
 
 		ret = check_int(ptr, &il);
 		if (ret)
 			break;
 
-		if (o->max_val && il > o->max_val)
-			il = o->max_val;
+		if (o->maxval && il > (int) o->maxval) {
+			fprintf(stderr, "max value out of range: %d (%d max)\n", il, o->maxval);
+			return 1;
+		}
+		if (o->minval && il < o->minval) {
+			fprintf(stderr, "min value out of range: %d (%d min)\n", il, o->minval);
+			return 1;
+		}
+
+		if (o->neg)
+			il = !il;
 
 		if (fn)
 			ret = fn(data, &il);
@@ -293,8 +363,11 @@ static int handle_option(struct fio_option *o, const char *ptr, void *data)
 	 * Do this before parsing the first round, to check if we should
 	 * copy set 1 options to set 2.
 	 */
-	if (ptr)
+	if (ptr && (o->type != FIO_OPT_STR_STORE)) {
 		ptr2 = strchr(ptr, ',');
+		if (!ptr2)
+			ptr2 = strchr(ptr, ':');
+	}
 
 	/*
 	 * Don't return early if parsing the first option fails - if
@@ -361,3 +434,80 @@ int parse_option(const char *opt, struct fio_option *options, void *data)
 	fprintf(stderr, "fio: failed parsing %s\n", opt);
 	return 1;
 }
+
+int show_cmd_help(struct fio_option *options, const char *name)
+{
+	int show_all = !strcmp(name, "all");
+	const char *typehelp[] = {
+		"string (opt=bla)",
+		"string with possible k/m/g postfix (opt=4k)",
+		"string with range and postfix (opt=1k-4k)",
+		"string with time postfix (opt=10s)",
+		"string (opt=bla)",
+		"string with dual range (opt=1k-4k,4k-8k)",
+		"integer value (opt=100)",
+		"boolean value (opt=1)",
+		"no argument (opt)",
+	};
+	struct fio_option *o;
+	int found = 0;
+
+	for (o = &options[0]; o->name; o++) {
+		int match = !strcmp(name, o->name);
+
+		if (show_all || match) {
+			found = 1;
+			printf("%20s: %s\n", o->name, o->help);
+			if (show_all)
+				continue;
+		}
+
+		if (!match)
+			continue;
+
+		printf("%20s: %s\n", "type", typehelp[o->type]);
+		printf("%20s: %s\n", "default", o->def ? o->def : "no default");
+		show_option_range(o);
+		show_option_values(o);
+	}
+
+	if (found)
+		return 0;
+
+	printf("No such command: %s\n", name);
+	return 1;
+}
+
+/*
+ * Handle parsing of default parameters.
+ */
+void fill_default_options(void *data, struct fio_option *options)
+{
+	struct fio_option *o;
+
+	for (o = &options[0]; o->name; o++)
+		if (o->def)
+			handle_option(o, o->def, data);
+}
+
+/*
+ * Sanitize the options structure. For now it just sets min/max for bool
+ * values and whether both callback and offsets are given.
+ */
+void options_init(struct fio_option *options)
+{
+	struct fio_option *o;
+
+	for (o = &options[0]; o->name; o++) {
+		if (o->type == FIO_OPT_BOOL) {
+			o->minval = 0;
+			o->maxval = 1;
+		}
+		if (!o->cb && !o->off1)
+			fprintf(stderr, "Option %s: neither cb nor offset given\n", o->name);
+		if (o->type == FIO_OPT_STR)
+			continue;
+		if (o->cb && (o->off1 || o->off2 || o->off3 || o->off4))
+			fprintf(stderr, "Option %s: both cb and offset given\n", o->name);
+	}
+}
