@@ -4,6 +4,7 @@
 #include "fio.h"
 
 static struct timeval genesis;
+static unsigned long ns_granularity;
 
 unsigned long utime_since(struct timeval *s, struct timeval *e)
 {
@@ -74,40 +75,56 @@ void __usec_sleep(unsigned int usec)
 
 void usec_sleep(struct thread_data *td, unsigned long usec)
 {
-	struct timespec req, rem;
-
-	req.tv_sec = usec / 1000000;
-	req.tv_nsec = usec * 1000 - req.tv_sec * 1000000;
+	struct timespec req;
+	struct timeval tv;
 
 	do {
-		if (usec < 5000) {
+		unsigned long ts = usec;
+
+		if (usec < ns_granularity) {
 			__usec_sleep(usec);
 			break;
 		}
 
-		rem.tv_sec = rem.tv_nsec = 0;
-		if (nanosleep(&req, &rem) < 0)
+		ts = usec - ns_granularity;
+
+		if (ts >= 1000000) {
+			req.tv_sec = ts / 1000000;
+			ts -= 1000000 * req.tv_sec;
+		} else
+			req.tv_sec = 0;
+
+		req.tv_nsec = ts * 1000;
+		fio_gettime(&tv, NULL);
+
+		if (nanosleep(&req, NULL) < 0)
 			break;
 
-		if ((rem.tv_sec + rem.tv_nsec) == 0)
+		ts = utime_since_now(&tv);
+		if (ts >= usec)
 			break;
 
-		req.tv_nsec = rem.tv_nsec;
-		req.tv_sec = rem.tv_sec;
-
-		usec = rem.tv_sec * 1000000 + rem.tv_nsec / 1000;
+		usec -= ts;
 	} while (!td->terminate);
 }
 
 void rate_throttle(struct thread_data *td, unsigned long time_spent,
-		   unsigned int bytes, int ddir)
+		   unsigned int bytes)
 {
 	unsigned long usec_cycle;
+	unsigned int bs;
 
 	if (!td->rate)
 		return;
 
-	usec_cycle = td->rate_usec_cycle * (bytes / td->min_bs[ddir]);
+	if (td_rw(td))
+		bs = td->rw_min_bs;
+	else if (td_read(td))
+		bs = td->min_bs[DDIR_READ];
+	else
+		bs = td->min_bs[DDIR_WRITE];
+
+	usec_cycle = td->rate_usec_cycle * (bytes / bs);
 
 	if (time_spent < usec_cycle) {
 		unsigned long s = usec_cycle - time_spent;
@@ -130,6 +147,30 @@ unsigned long mtime_since_genesis(void)
 }
 
 static void fio_init time_init(void)
+{
+	int i;
+
+	/*
+	 * Check the granularity of the nanosleep function
+	 */
+	for (i = 0; i < 10; i++) {
+		struct timeval tv;
+		struct timespec ts;
+		unsigned long elapsed;
+
+		fio_gettime(&tv, NULL);
+		ts.tv_sec = 0;
+		ts.tv_nsec = 1000;
+
+		nanosleep(&ts, NULL);
+		elapsed = utime_since_now(&tv);
+
+		if (elapsed > ns_granularity)
+			ns_granularity = elapsed;
+	}
+}
+
+void set_genesis_time(void)
 {
 	fio_gettime(&genesis, NULL);
 }
